<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="krdc" version="10">
<MenuBar>
    <Menu name="file">
        <Action name="new_connection" />
    </Menu>
    <Menu noMerge="1" name="session"><text>&amp;Session</text>
        <Action name="switch_fullscreen" />
        <Action name="take_screenshot" />
        <Action name="view_only" />
        <Action name="clipboard_sharing" />
        <Action name="show_local_cursor" />
        <Action name="grab_all_keys" />
        <Action name="scale" />
        <Action name="disconnect" />
    </Menu>
    <Action name="bookmark" />
    <Menu name="settings">
        <Action name="remote_desktop_dockwidget" />
    </Menu>
</MenuBar>
<ToolBar fullWidth="false" name="krdc_remote_view_toolbar" newline="false"><text>Remote View Toolbar</text>
    <Action name="new_connection" />
    <Separator />
    <Action name="switch_fullscreen" />
    <Action name="take_screenshot" />
    <Action name="view_only" />
    <Action name="clipboard_sharing" />
    <Action name="show_local_cursor" />
    <Action name="grab_all_keys" />
    <Action name="scale" />
    <Action name="disconnect" />
</ToolBar>
</gui>
